from enum import Enum
from typing import Any, Dict


class ErrorType(str, Enum):
    INVALID_FIELD_EXTRA = """Cannot get value for %(field)s: %(msg)s.

    Please use the following pattern:

        Field(..., %(extra_field)s={
            "provider": {
                "instance": provider_instance,
                "options": {...},
            }
        })

    Skipping.
    """

    INVALID_PROVIDER_INSTANCE = (
        "Given provider instance is not an instance of Provider interface."
    )

    PROVIDER_WAS_NOT_SPECIFIED = "No provider was specified. Skipping."


class ProviderError(Exception):
    def __init__(self, error_type: ErrorType, kwargs: Dict[str, Any] = {}) -> None:
        self.error_msg = error_type.value % kwargs
        self.error_type = error_type
        super().__init__(self.error_msg)



import logging
from typing import Any, Dict, Optional, Union

from pydantic_settings_external.exceptions import ErrorType, ProviderError
from pydantic_settings_external.types import Provider, ProviderDict
from pydantic_settings_external.version import PYDANTIC_MAJOR_VERSION

logger = logging.getLogger(__package__)


def validate_provider_field(
    name: str, field_extra: Dict[str, Any] | None = None
) -> ProviderDict:
    pydantic_extra_field_name = (
        "extra" if PYDANTIC_MAJOR_VERSION == "1" else "json_field_extra"
    )

    if field_extra is None:
        raise ProviderError(ErrorType.PROVIDER_WAS_NOT_SPECIFIED)

    if "provider" not in field_extra:
        raise ProviderError(ErrorType.PROVIDER_WAS_NOT_SPECIFIED)

    provider: ProviderDict = field_extra["provider"]

    if "instance" not in provider:
        raise ProviderError(
            ErrorType.INVALID_FIELD_EXTRA,
            {
                "msg": f"key 'instance' not in {pydantic_extra_field_name}['provider']",
                "extra_field": pydantic_extra_field_name,
                "field": name,
            },
        )

    if "options" not in provider:
        raise ProviderError(
            ErrorType.INVALID_FIELD_EXTRA,
            {
                "msg": f"key 'options' not in {pydantic_extra_field_name}",
                "extra_field": pydantic_extra_field_name,
                "field": name,
            },
        )

    if not isinstance(provider["options"], dict):
        raise ProviderError(
            ErrorType.INVALID_FIELD_EXTRA,
            {
                "msg": f"key {pydantic_extra_field_name}['options'] should be a dict",
                "extra_field": pydantic_extra_field_name,
                "field": name,
            },
        )

    provider_instance = provider["instance"]

    if not isinstance(provider_instance, Provider):
        raise ProviderError(ErrorType.INVALID_PROVIDER_INSTANCE)

    return provider


def log_error_msg(exc: ProviderError):
    logger.warning(exc)


def display_hint(hint: str):
    logger.info(hint)


def get_field_value(
    name: str, field_extra: Optional[Dict[str, Any]] = None
) -> Union[str, None]:
    provider = validate_provider_field(name, field_extra)
    provider_instance = provider["instance"]
    provider_options = provider["options"]
    provider_hint = provider.get("hint")

    value = provider_instance.get(options=provider_options)

    if value is None and provider_hint is not None:
        display_hint(provider_hint)

    return value



from typing import Type

from pydantic import BaseModel

from pydantic_settings_external.version import PYDANTIC_MAJOR_VERSION, PYDANTIC_VERSION

BaseSettings: Type[BaseModel]

if PYDANTIC_MAJOR_VERSION == "1":
    from pydantic_settings_external.pydantic.v1 import BaseSettings as _BaseSettings

    BaseSettings = _BaseSettings
elif PYDANTIC_MAJOR_VERSION == "2":
    from pydantic_settings_external.pydantic.v2 import BaseSettings as _BaseSettings

    BaseSettings = _BaseSettings
else:
    raise ImportError(
        f"Installed version {PYDANTIC_VERSION} of PyDantic is not supporting."
    )

__all__ = ("BaseSettings",)



from pydantic_settings_external.version import PYDANTIC_MAJOR_VERSION

if PYDANTIC_MAJOR_VERSION != "1":
    raise ImportError("This package required Pydantic v1")

from typing import Any, Dict, Tuple

from pydantic import BaseSettings as PydanticBaseSettings
from pydantic.env_settings import SettingsSourceCallable

from pydantic_settings_external.exceptions import ErrorType, ProviderError
from pydantic_settings_external.utils import get_field_value, log_error_msg


def external_settings_source(settings: PydanticBaseSettings) -> Dict[str, Any]:  # C901
    d: Dict[str, Any] = {}

    for field in settings.__fields__.values():
        try:
            d[field.alias] = get_field_value(field.name, field.field_info.extra)
        except ProviderError as exc:
            if exc.error_type is not ErrorType.PROVIDER_WAS_NOT_SPECIFIED:
                log_error_msg(exc)

            if exc.error_type is ErrorType.INVALID_PROVIDER_INSTANCE:
                raise exc

            continue

    return d


class BaseSettings(PydanticBaseSettings):
    class Config:
        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            return (
                init_settings,
                env_settings,
                file_secret_settings,
                external_settings_source,
            )



from typing import Any, Dict, Optional

import pytest
from pydantic import Field

from pydantic_settings_external import BaseSettings
from pydantic_settings_external.version import PYDANTIC_MAJOR_VERSION


@pytest.mark.skipif(
    PYDANTIC_MAJOR_VERSION != "2", reason="requires Pydantic 2.0 or higher."
)
def test_pydantic_v2():
    class FakeProvider:
        def get(self, options: Dict[str, Any]) -> Optional[Any]:
            return "123"

    fake_provider = FakeProvider()

    class Settings(BaseSettings):
        SENDGRID_API_KEY: str = Field(
            ...,
            json_schema_extra={
                "provider": {
                    "instance": fake_provider,
                    "options": {"name": "sendgrid-api-key", "vesion": "latest"},
                    "hint": "",
                }
            },
        )

    settings = Settings()
    assert settings.SENDGRID_API_KEY == "123"


@pytest.mark.skipif(
    PYDANTIC_MAJOR_VERSION != "2", reason="requires Pydantic 2.0 or higher."
)
def test_pydantic_v2_dotenv(monkeypatch):
    monkeypatch.setenv("SENDGRID_API_KEY", "444")

    class FakeProvider:
        def get(self, options: Dict[str, Any]) -> Optional[Any]:
            return None

    fake_provider = FakeProvider()

    class Settings(BaseSettings):
        SENDGRID_API_KEY: str = Field(
            ...,
            json_schema_extra={
                "provider": {
                    "instance": fake_provider,
                    "options": {"name": "sendgrid-api-key", "vesion": "latest"},
                    "hint": "",
                }
            },
        )

    settings = Settings()

    assert settings.SENDGRID_API_KEY == "444"



from pydantic_settings_external.version import PYDANTIC_MAJOR_VERSION

if PYDANTIC_MAJOR_VERSION != "2":
    raise ImportError("This package required Pydantic v2")

from pydantic.fields import FieldInfo

from pydantic_settings_external.exceptions import ErrorType, ProviderError
from pydantic_settings_external.utils import get_field_value, log_error_msg

try:
    from pydantic_settings.main import BaseSettings as PyDanticBaseSettings
    from pydantic_settings.sources import PydanticBaseSettingsSource
except ImportError:
    raise ImportError(
        "In order to use this package for Pydantic v2 it is required to install "
        "pydantic-settings package. "
        "Please install it using the following command: pip install pydantic-settings."
    )

from typing import Any, Dict, Tuple, Type


class ExternalSettingsSource(PydanticBaseSettingsSource):
    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> Tuple[Any, str, bool]:
        return get_field_value(field_name, field.json_schema_extra), field_name, False

    def __call__(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {}

        for field_name, field in self.settings_cls.model_fields.items():
            try:
                field_value, field_key, value_is_complex = self.get_field_value(
                    field, field_name
                )
            except ProviderError as exc:
                if exc.error_type is not ErrorType.PROVIDER_WAS_NOT_SPECIFIED:
                    log_error_msg(exc)

                if exc.error_type is ErrorType.INVALID_PROVIDER_INSTANCE:
                    raise exc
                continue

            field_value = self.prepare_field_value(
                field_name, field, field_value, value_is_complex
            )

            if field_value is not None:
                d[field_key] = field_value

        return d


class BaseSettings(PyDanticBaseSettings):
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[PyDanticBaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            file_secret_settings,
            ExternalSettingsSource(settings_cls),
        )



try:
    from pydantic import VERSION as PYDANTIC_VERSION
except ImportError:
    raise ImportError(
        "Neither the 1.10.x nor the 2.0.x version of Pydantic was installed. "
        "Please install one of the required versions of Pydantic."
    )

VERSION = "0.1.0"

PYDANTIC_MAJOR_VERSION = PYDANTIC_VERSION.split(".")[0]



from typing import Any, Dict, Optional, Protocol, Tuple, TypedDict, runtime_checkable


@runtime_checkable
class Provider(Protocol):
    def get(self, options: Dict[str, Any]) -> Optional[Any]:
        """Returns the value from external provider by key.

        Args:
            options (Dict[str, Any]): key specific options for provider.

        Returns:
            Optional[str]: the value of the key from current provider.
        """


OptionsType = Dict[str, Any]


class ProviderDict(TypedDict):
    instance: Provider
    options: OptionsType
    hint: Optional[str]


ProviderOptions = Tuple[Provider, OptionsType]
""" Type of providers registry. """
